import logging
import os
from datetime import datetime
from pathlib import Path

from rich.console import Console
from rich.logging import RichHandler
from rich.pretty import install
from rich.traceback import install as ins


def setup_logging(name) -> logging.Logger:
    """Set's up logging"""
    
    install()
    ins()
    
    console = Console(record=True, force_terminal=True)
    
    path = Path(os.getcwd()) / "logs"
    path.mkdir(exist_ok=True)
    file_handler = logging.FileHandler(
        filename=f"logs/{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}.log",
        mode="w",
        encoding="utf-8",
    )
    file_handler.setFormatter(logging.Formatter("%(message)s"))
    file_handler.setLevel(logging.INFO)
    console_handler = RichHandler(
        console=console,
        show_time=True,
        show_path=True,
        markup=True,
        rich_tracebacks=True,
        tracebacks_show_locals=True,
        tracebacks_extra_lines=5,
        tracebacks_theme="solarized",
    )
    console_handler.setFormatter(logging.Formatter("%(message)s"))
    console_handler.setLevel(logging.INFO)
    logging.basicConfig(level=logging.INFO, handlers=[file_handler, console_handler])
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    
    return logger



from __future__ import annotations

import asyncio
import os
from collections import defaultdict
from datetime import datetime
from typing import Any, List, Optional, Type, TypeVar

from pydantic.main import ModelMetaclass

from .logging import setup_logging

try:
    import query as q

except ImportError:
    from . import query as q

from dotenv import load_dotenv
from pydantic import BaseModel  # pylint: disable=no-name-in-module

from .client import FaunaClient
from .errors import FaunaException
from .json import JSONModel

load_dotenv()

T = TypeVar("T", bound="FaunaModel")



class FaunaModelMetaclass(ModelMetaclass):
    def __new__(cls, name, bases, attrs):
        new_cls = super().__new__(cls, name, bases, attrs)
        cls.Metadata.__subclasses__.append(new_cls)
        return new_cls

    class Metadata:
        __subclasses__: List[Type[FaunaModel]] = []


class FaunaModel(JSONModel, metaclass=FaunaModelMetaclass):
    ref: Optional[str] = None

    ts: Optional[str] = None

    def __init__(self, **data: Any):
        super().__init__(**data)

        for field in self.__fields__.values():
            try:
                one_of = field.field_info.extra.get("oneOf")

                if isinstance(one_of, list):
                    if data.get(field.name) not in one_of:
                        raise ValueError(f"{field.name} must be one of {one_of}")

            except Exception:
                continue

    @classmethod
    @property
    def logger(cls):
        return setup_logging(cls.__name__)

    @classmethod
    async def create_all(cls):
        await asyncio.gather(
            *[model.provision() for model in cls.Metadata.__subclasses__]
        )

    @classmethod
    def client(cls):
        return FaunaClient()

    @classmethod
    def q(cls):
        return cls.client().query

    @classmethod
    async def provision(cls) -> bool:
        _q = cls.q()

        try:
            if not await _q(q.exists(q.collection(cls.__name__.lower()))):
                await _q(q.create_collection({"name": cls.__name__.lower()}))

                cls.logger.info("Created collection %s", cls.__name__.lower())

            if not await _q(q.exists(q.index(cls.__name__.lower()))):
                await _q(
                    q.create_index(
                        {
                            "name": cls.__name__.lower(),
                            "source": q.collection(cls.__name__.lower()),
                        }
                    )
                )

                cls.logger.info("Created index %s", cls.__name__.lower())

            for field in cls.__fields__.values():
                if field.field_info.extra.get("unique"):
                    if not await _q(
                        q.exists(q.index(f"{cls.__name__.lower()}_{field.name}_unique"))
                    ):
                        await _q(
                            q.create_index(
                                {
                                    "name": f"{cls.__name__.lower()}_{field.name}_unique",
                                    "source": q.collection(cls.__name__.lower()),
                                    "terms": [{"field": ["data", field.name]}],
                                    "unique": True,
                                }
                            )
                        )

                        cls.logger.info(
                            "Created unique index %s_%s",
                            cls.__name__.lower(),
                            field.name,
                        )
                    continue

                if field.field_info.extra.get("index"):
                    if not await _q(
                        q.exists(q.index(f"{cls.__name__.lower()}_{field.name}"))
                    ):
                        await _q(
                            q.create_index(
                                {
                                    "name": f"{cls.__name__.lower()}_{field.name}",
                                    "source": q.collection(cls.__name__.lower()),
                                    "terms": [{"field": ["data", field.name]}],
                                }
                            )
                        )

                        cls.logger.info(
                            "Created index %s_%s", cls.__name__.lower(), field.name
                        )
                        continue

            return True

        except (FaunaException, KeyError, TypeError) as exc:
            cls.logger.error(exc)

            return False

    @classmethod
    async def find_unique(cls: Type[T], field: str, value: Any) -> Optional[T]:
        try:
            data = await cls.q()(
                q.get(q.match(q.index(f"{cls.__name__.lower()}_{field}_unique"), value))
            )
            return cls(
                **{
                    **data["data"],  # type: ignore
                    "ref": data["ref"]["@ref"]["id"],  # type: ignore
                    "ts": data["ts"] / 1000,  # type: ignore
                }
            )

        except (FaunaException, KeyError, TypeError) as exc:
            cls.logger.error(exc)
            return None

    @classmethod
    async def find_many(cls: Type[T], field: str, value: Any) -> List[T]:
        try:
            _q = cls.q()

            refs = await _q(
                q.paginate(q.match(q.index(f"{cls.__name__.lower()}_{field}"), value))
            )

            assert isinstance(refs, dict)

            refs = refs["data"]

            collection_refs_map = defaultdict(list)

            for ref in refs:
                collection = q.collection(cls.__name__.lower())

                collection_refs_map[collection].append(
                    q.ref(collection, ref["@ref"]["id"])  # type: ignore
                )

            data = await asyncio.gather(
                *[
                    _q(q.get(q.paginate(collection_refs)))
                    for _, collection_refs in collection_refs_map.items()
                ]
            )

            flattened_data = [
                item for collection_data in data for item in collection_data["data"]
            ]

            return [
                cls(
                    **{**d["data"], "ref": d["ref"]["@ref"]["id"], "ts": d["ts"] / 1000}
                )
                for d in flattened_data
            ]

        except (FaunaException, KeyError, TypeError) as exc:
            cls.logger.error(exc)

            return []

    @classmethod
    async def get(cls: Type[T], ref: str) -> Optional[T]:
        try:
            data = await cls.q()(q.get(q.ref(q.collection(cls.__name__.lower()), ref)))
            return cls(
                **{
                    **data["data"],  # type: ignore
                    "ref": data["ref"]["@ref"]["id"],  # type: ignore
                    "ts": data["ts"] / 1000,  # type: ignore
                }
            )

        except (FaunaException, KeyError, TypeError) as exc:
            cls.logger.error(exc)
            return None

    @classmethod
    async def all(cls: Type[T]) -> List[T]:
        try:
            _q = cls.q()

            query = q.paginate(q.match(q.index(f"{cls.__name__.lower()}")))

            refs = (await _q(query))["data"]  # type: ignore

            data = await asyncio.gather(
                *[
                    _q(
                        q.get(
                            q.ref(q.collection(cls.__name__.lower()), ref["@ref"]["id"])
                        )
                    )
                    for ref in refs
                ]
            )

            return [
                cls(
                    **{**d["data"], "ref": d["ref"]["@ref"]["id"], "ts": d["ts"] / 1000}
                )
                for d in data
            ]

        except (FaunaException, KeyError, TypeError) as exc:
            cls.logger.error(exc)

            return []

    @classmethod
    async def delete_one(cls, field: str, value: Any) -> bool:
        try:
            _q = cls.q()

            ref = await _q(
                q.get(q.match(q.index(f"{cls.__name__.lower()}_{field}_unique"), value))
            )

            await _q(q.delete(ref))

            return True

        except (FaunaException, KeyError, TypeError) as exc:
            cls.logger.error(exc)

            return False

    @classmethod
    async def delete(cls, ref: str) -> bool:
        try:
            await cls.q()(q.delete(q.ref(q.collection(cls.__name__.lower()), ref)))

            return True

        except (FaunaException, KeyError, TypeError) as exc:
            cls.logger.error(exc)

            return False

    async def create(self: T) -> Optional[T]:
        try:
            for field in self.__fields__.values():
                if field.field_info.extra.get("unique"):
                    instance = await self.find_unique(
                        field.name, self.__dict__[field.name]
                    )

                    if instance is None:
                        continue

                    if issubclass(instance.__class__, FaunaModel):
                        return instance

            data = await self.__class__.q()(
                q.create(
                    q.collection(self.__class__.__name__.lower()), {"data": self.dict()}
                )
            )

            self.ref = data["ref"]["@ref"]["id"]  # type: ignore

            self.ts = data["ts"] / 1000  # type: ignore
            return self

        except (FaunaException, KeyError, TypeError) as exc:
            self.logger.error(exc)
            return None

    @classmethod
    async def update(cls: Type[T], ref: str, **kwargs) -> T:
        try:
            instance = await cls.get(ref)

            if isinstance(instance, cls):
                instance_updated = await cls.q()(
                    q.update(
                        q.ref(q.collection(cls.__name__.lower()), ref),
                        {"data": kwargs.get("kwargs", kwargs)},
                    )
                )
                return cls(
                    **{
                        **instance_updated["data"],  # type: ignore
                        "ref": instance_updated["ref"]["@ref"]["id"],  # type: ignore
                        "ts": instance_updated["ts"] / 1000,  # type: ignore
                    }
                )

            else:
                raise ValueError(f"Field {ref} not found")

        except (FaunaException, KeyError, TypeError) as exc:
            cls.logger.error(exc)

            raise ValueError(f"Field {ref} not found")  # pylint: disable=all

    async def save(self: T) -> Optional[T]:
        if isinstance(self.ref, str) and len(self.ref) == 18:
            return await self.update(self.ref, kwargs=self.dict())

        return await self.create()



import asyncio
import logging
from typing import Optional, cast

from aiohttp.web import Application, AppRunner, TCPSite
from rich.console import Console
from rich.logging import RichHandler
from watchdog.events import (
    FileCreatedEvent,
    FileDeletedEvent,
    FileModifiedEvent,
    FileSystemEventHandler,
)
from watchdog.observers import Observer

from .api import Api

console = Console(record=True)


class ReloadHandler(FileSystemEventHandler):
    def __init__(self, loop, task, logger):
        self.loop = loop
        self.task = task
        self.logger = logger

    def on_modified(self, event):
        self.logger.info(f"File modified: {event.src_path}")
        self.loop.call_soon_threadsafe(self.task.cancel)

    def on_created(self, event):
        self.logger.info(f"File created: {event.src_path}")
        self.loop.call_soon_threadsafe(self.task.cancel)

    def on_deleted(self, event):
        self.logger.info(f"File deleted: {event.src_path}")
        self.loop.call_soon_threadsafe(self.task.cancel)


def setup_logging():
    logging.basicConfig(
        level=logging.INFO,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler(console=console, rich_tracebacks=True)],
        force=True,
    )

    logger = logging.getLogger("rich")
    logger.setLevel(logging.INFO)
    return logger


class AioFauna(Api):
    """AioFauna CLI"""

    runner: Optional[AppRunner] = None
    site: Optional[TCPSite] = None
    file_watcher: Optional[asyncio.Task] = None
    host: str = "localhost"
    port: int = 8080

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = setup_logging()

    async def run_(self):
        self.runner = AppRunner(cast(Application, self))
        self.logger.info("Starting server...")
        await self.runner.setup()
        self.site = TCPSite(self.runner, self.host, self.port)
        await self.site.start()
        self.logger.info("Server running on http://%s:%s", self.host, self.port)

    def run(self):
        loop = asyncio.get_event_loop()
        while True:
            task = asyncio.ensure_future(self.run_(), loop=loop)
            observer = Observer()
            observer.schedule(
                ReloadHandler(loop, task, self.logger), ".", recursive=True
            )
            observer.start()

            try:
                loop.run_forever()
            except KeyboardInterrupt:
                observer.stop()
                observer.join()
                self.logger.info("Stopping server...")
                break
            except Exception as e:
                observer.stop()
                observer.join()
                self.logger.error(e)
                break



import base64
import functools
import io
import json
import logging
import os
from dataclasses import dataclass, field
from re import T
from typing import Any, AsyncGenerator, Dict, List, Literal, Optional, Union

from aiohttp import (ClientConnectionError, ClientConnectorSSLError,
                     ClientResponse, ClientSession, ClientTimeout,
                     TCPConnector)
from aiohttp.web_exceptions import HTTPException
from dotenv import load_dotenv
from multidict import CIMultiDict
from pydantic import BaseModel

from .errors import FaunaException
from .json import FaunaJSONEncoder, to_json
from .logging import setup_logging
from .objects import Expr
from .typedefs import LazyProxy

load_dotenv()

logging_factory = functools.partial(setup_logging, __name__)

Method = Literal["GET", "POST", "PUT", "PATCH", "DELETE"]
Json = Union[Dict[str, Any], List[Dict[str, Any]]]
MaybeJson = Optional[Json]
Headers = Dict[str, str]
MaybeHeaders = Optional[Headers]

# Load environment variables once at the start of the program
FAUNA_SECRET = os.environ["FAUNA_SECRET"]
HEADERS = {
    "Authorization": f"Bearer {FAUNA_SECRET}",
    "Content-type": "application/json",
    "Accept": "application/json"
}


@dataclass(init=True, repr=True, unsafe_hash=False, frozen=False)
class FaunaClient(LazyProxy[ClientSession]):
    secret: str = field(default=FAUNA_SECRET)
    logger:logging.Logger = field(default_factory=logging_factory)
    session: Optional[ClientSession] = field(default=None)
        
    def __load__(self) -> ClientSession:
        if self.session is None:
            return ClientSession(
                "https://db.fauna.com",
                headers=CIMultiDict(HEADERS),
                response_class=ClientResponse,
                connector=TCPConnector(ssl=False, limit=1000,
                keepalive_timeout=10),
                timeout=ClientTimeout(total=10),
                connector_owner=True,
                trust_env=True,
                read_bufsize=2**18,
            )
        else:
            return self.session
    
    async def query(self, expr: Expr,session: Optional[ClientSession]=None) -> Any:
        if session is None:
            session = self.__load__()
        async with session as session:
            async with session.post(
                "/",
                data=to_json(expr),
            ) as response:
                try:
                    data = await response.json()
                    self.logger.info(data)
                    if data.get("resource") is not None:
                        return data["resource"]
                    if data.get("error") is not None:
                        return data["error"]
                    return data

                except (
                        FaunaException,
                        ValueError,
                        KeyError,
                        TypeError,
                        Exception,
                        UnicodeError,
                        json.JSONDecodeError,
                        RuntimeError,
                        ClientConnectionError,
                        ClientConnectorSSLError
                ) as exc:  # pylint:disable=all
                    self.logger.error(exc)
                    return {"errors": exc}
                finally:
                    await session.close()
    
    
    async def stream(self, expr: Expr) -> AsyncGenerator[str, None]:
        async with self.__load__() as session:
            async with session.post(
                "",
                data=to_json(expr),
                headers={
                    "Authorization": f"Bearer {self.secret}",
                    "Content-type": "application/json",
                    "Accept": "text/event-stream",
                    "Keep-Alive": "timeout=5, max=900",
                    "Connection": "keep-alive",
                    "Cache-Control": "no-cache",
                    "X-Last-Seen-Txn": "0",
                    "X-Request-By": "aiofauna",
                    "X-Query-By": "@obahamonde",
                },
            ) as response:
                async for chunk in response.content.iter_any():
                    try:
                        yield chunk.decode()
                    except (
                        HTTPException,
                        FaunaException,
                        ValueError,
                        KeyError,
                        TypeError,
                        Exception,
                        UnicodeError,
                        json.JSONDecodeError,
                        RuntimeError,
                        ClientConnectionError,
                        ClientConnectorSSLError
                    ) as exc:
                        self.logger.error(exc)
                        yield json.dumps({"errors": exc})


@dataclass(init=True, repr=True, unsafe_hash=False, frozen=False)
class ApiClient(LazyProxy[ClientSession]):
    """
    Generic HTTP Client
    """
    base_url: Optional[str] = field(default=None)
    headers: Optional[Headers] = field(default=None)
    limit: int = field(default=1000)
    _session: Optional[ClientSession] = field(default=None)
    
    def __post_init__(self):
        self.logger = setup_logging(self.__class__.__name__)
        if not ApiClient._session:
            ApiClient._session = self.__load__()                
    
    
    def __load__(self) -> ClientSession:
        if self.base_url and self.headers:
            return ClientSession(
                self.base_url,
                headers=CIMultiDict(self.headers),
                response_class=ClientResponse,
                connector=TCPConnector(ssl=False, limit=self.limit,
                keepalive_timeout=10),
                timeout=ClientTimeout(total=10),
                connector_owner=True,
                trust_env=True,
                read_bufsize=2**16,
            )
        elif self.base_url:
            return ClientSession(
                self.base_url,
                response_class=ClientResponse,
                connector=TCPConnector(ssl=False, limit=self.limit,
                keepalive_timeout=10),
                timeout=ClientTimeout(total=10),
                connector_owner=True,
                trust_env=True,
                read_bufsize=2**16,
            )
        elif self.headers:
            return ClientSession(
                headers=CIMultiDict(self.headers),
                response_class=ClientResponse,
                connector=TCPConnector(ssl=False, limit=self.limit,
                keepalive_timeout=10),
                timeout=ClientTimeout(total=10),
                connector_owner=True,
                trust_env=True,
                read_bufsize=2**16,
            )
        else:
            return ClientSession(
                response_class=ClientResponse,
                connector=TCPConnector(ssl=False, limit=self.limit,
                keepalive_timeout=10),
                timeout=ClientTimeout(total=10),
                connector_owner=True,
                trust_env=True,
                read_bufsize=2**16,
            )
            

    async def fetch(
        self,
        url: str,
        method: Method = "GET",
        headers: MaybeHeaders = None,
        json: MaybeJson = None):
        if self.base_url is not None:
            url = self.base_url + url
        if self.headers is not None and headers is not None:
            headers = {**self.headers, **headers}
        elif self.headers is not None:
            headers = self.headers
        async with self.__load__() as session:
            async with session.request(
                method, url, headers=headers, json=json, timeout=30
            ) as response:
                try:
                    data = await response.json()
                    return data
                except (
                    HTTPException,
                    FaunaException,
                    ValueError,
                    KeyError,
                    TypeError,
                    Exception,
                ) as exc:  # pylint:disable=broad-exception-caught, unused-variable
                    print(exc)
                    return None

    async def text(
        self,
        url: str,
        method: Method = "GET",
        headers: MaybeHeaders = None,
        json: MaybeJson = None,
    ):
        if self.base_url is not None:
            url = self.base_url + url
        if self.headers is not None and headers is not None:
            headers = {**self.headers, **headers}
        elif self.headers is not None:
            headers = self.headers
        async with self.__load__() as session:
            async with session.request(
                method, url, headers=headers, json=json, timeout=30
            ) as response:
                try:
                    data = await response.text()
                    return data
                except (
                    HTTPException,
                    FaunaException,
                    ValueError,
                    KeyError,
                    TypeError,
                    Exception,
                ) as exc:  # pylint:disable=broad-exception-caught, unused-variable
                    print(exc)
                    return None  # type: ignore

    async def stream(
        self,
        url: str,
        method: Method = "GET",
        headers: MaybeHeaders = None,
        json: MaybeJson = None,
        session: Optional[ClientSession] = None,
    ):
        if self.base_url is not None:
            url = self.base_url + url
        if self.headers is not None and headers is not None:
            headers = {**self.headers, **headers}
        elif self.headers is not None:
            headers = self.headers
        async with self.__load__() as session:
            async with session.request(
                method, url, headers=headers, json=json, timeout=30
            ) as response:
                async for chunk in response.content.iter_any():
                    try:
                        yield chunk.decode()
                    except (
                        HTTPException,
                        FaunaException,
                        ValueError,
                        KeyError,
                        TypeError,
                        Exception,
                    ) as exc:
                        print(exc)